//
//          CMVSource 1.0.0 for AviSynth 2.5.x
//
//  Loads Dwarf Fortress CMV and CCMV movies. Latest release hosted at
//  http://www.gyroshot.com/cmvsource.htm
//
//          Copyright 2011 Robert Martens  robert.martens@gmail.com
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software Foundation,
//  Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.



#include "CmvSource.h"

#include <cstring>  // memcpy, memset
#include <string>
#include <vector>
#include <Windows.h>

#include "avisynth.h"
#include "DfUtils.h"
#include "zconf.h"
#include "zlib.h"



CmvSource::CmvSource( std::vector<char> _cmvBuf, PClip _tileset,
                      std::string _colors, int _tileW, int _tileH,
                      IScriptEnvironment* env ) :
                      colors(LoadColors(_colors, env)),
                      set(_tileset->GetFrame(0, env)),
                      tileW(_tileW), tileH(_tileH), zlibData(true)

{
  
  // I feel like an idiot for not having a better approach here, but I need to
  // read integers from a vector of characters, and memcpy seemed simplest.
  int version = 0;
  std::memcpy(&version, &_cmvBuf[0], 4);
  std::memcpy(&cmvCols, &_cmvBuf[4], 4);
  std::memcpy(&cmvRows, &_cmvBuf[8], 4);
  
  // As pointed out in the readme, there are several possible results when
  // saving CCMV files from the CMV Editor, and I've yet to find a reliable
  // way to deal with all of them properly. Until I do, stick with CMVs.
  if ((version & 0xFFFF) == 0x9C78)
    env->ThrowError("CMVSource: CCMVs from the CMV Editor are currently "
                    "unsupported! See CMVSource.txt for details.");

  // This is not technically airtight. The old, 2004-era prerelease CMV format
  // stores its frame count in the first field, instead of a version number. If
  // there were to ever surface a CMV in said format that had a frame count
  // matching any of these version numbers, this technique would fail. Luckily,
  // all the old movies Toady has yet provided are 1000 frames long, save a
  // single one that runs for 64 frames, so this covers everything.
  if (  version != 10000 && // 2005 prerelease
        version != 10001 && // Current version number as of 31.25
        version != 10005  ) // CMV output from CMV Editor
    zlibData = false;

  int delayRate = 0, soundCount = 0;

  if (zlibData) {
    std::memcpy(&delayRate, &_cmvBuf[12], 4);
    if (version == 10001)
      std::memcpy(&soundCount, &_cmvBuf[16], 4);
  } else {
    vi.num_frames = version;
  }
  
  int sndNamesSize = 50 * soundCount,
      sndTimingsSize = version == 10001 ? 12800 : 0,
      sndCountSize = version == 10001 ? 4 : 0;

  int hdrSize = zlibData ? 16 : 12;

  int chunkOffset = hdrSize + sndCountSize + sndNamesSize + sndTimingsSize;
  
  // Need this for the upcoming frameVec/attribVec creation loop, otherwise I'd
  // do it down by the other vi.num_frames assignment.
  frameSize = cmvCols * cmvRows;

  if (zlibData) {

    // Starting at chunkOffset, modern CMV/CCMV files have several individually
    // compressed chunks of frame data, one after the other.
    for ( unsigned int i = chunkOffset, subchunkSize = 0;
          i < _cmvBuf.size();
          i += subchunkSize ) {
      
      // Read the size of the upcoming chunk, then skip over the integer...
      std::memcpy(&subchunkSize, &_cmvBuf[i], 4);
      i += 4;

      // ...extract the chunk from the master file data...
      std::vector<unsigned char> subchunk;
      for (unsigned int j = 0; j < subchunkSize; ++j) {
        subchunk.push_back(_cmvBuf[i+j]);
      }
      
      // ...prep for decompression...
      z_stream chnkStrm;
      chnkStrm.zalloc = Z_NULL;
      chnkStrm.zfree = Z_NULL;
      chnkStrm.opaque = Z_NULL;
      chnkStrm.avail_in = subchunk.size();
      chnkStrm.next_in = &subchunk[0];
      int ret = inflateInit(&chnkStrm);
      if (ret != Z_OK)
        env->ThrowError("CMVSource: Couldn't initialize decompression! inflateInit() error %d", ret);
      
      const int step = 65536;
      std::vector<unsigned char> tmp(step);
      
      // ...and decompress the subchunk until all input's consumed.
      while (ret != Z_STREAM_END) {
        chnkStrm.avail_out = tmp.size() - chnkStrm.total_out;
        chnkStrm.next_out = &tmp[chnkStrm.total_out];
        ret = inflate(&chnkStrm, Z_FINISH);
        if ( !(ret == Z_BUF_ERROR || ret == Z_STREAM_END) )
          env->ThrowError("CMVSource: Couldn't decompress frames! inflate() error %d", ret);
        if (ret == Z_BUF_ERROR) {
          tmp.resize(tmp.size() + step);
        }
      }

      inflateEnd(&chnkStrm);

      // With the hard part out of the way, it's just a matter of loading the
      // two vectors appropriately. Frame data in the current CMV format is
      // stored as what I'd call "planar", with one big block of bytes
      // describing the tile indices for one frame, followed by another block
      // for the attributes, then the next frame, its attributes, and so on.
      for (unsigned long j = 0; j < chnkStrm.total_out; j += frameSize * 2) {
        for (int k = 0; k < frameSize; ++k) {
          int frameOffset = j + k;
          frameVec.push_back(tmp[frameOffset]);
          attribVec.push_back(tmp[frameOffset+frameSize]);
        }
      }

    }

    vi.num_frames = frameVec.size() / frameSize;

  } else {

    vi.num_frames = version;

    // The prerelease files have only one big chunk of frame data, and it's not
    // compressed, so reading it is simple. Instead of being planar, however,
    // these are interleaved, with one byte for tile data, then another for the
    // foreground color, background color, and finally foreground intensity.
    for ( int i = chunkOffset;
          i < chunkOffset + (vi.num_frames * frameSize * 4);
          i += 4 ) {

      frameVec.push_back(_cmvBuf[i]);

      unsigned char fgColor = _cmvBuf[i+1],
                    bgColor = _cmvBuf[i+2];
      
      // Intensity for the old CMV format seems to be stored as a three state
      // byte, with values of 0, 1, or 128, and I'm not sure I know why. Here
      // I just treat it as zero or non-zero, and get seemingly proper results.
      unsigned char intense = _cmvBuf[i+3] ? 1 : 0;

      // In order to eliminate an ugly branch in GetFrame, I conform the tile
      // attributes for these old format CMVs to the new format, packing the
      // three bytes into one, the bits arranged as UIBGRBGR: Unused, Intense,
      // background Blue, Green, Red, foreground Blue, Green, Red.
      attribVec.push_back(intense << 6 | bgColor << 3 | fgColor);
      
    }

  }
  
  vi.width = cmvCols * tileW;
  vi.height = cmvRows * tileH;
  vi.fps_numerator = 100;
  vi.fps_denominator = delayRate > 0 ? delayRate : 1;
  vi.pixel_type = VideoInfo::CS_BGR32;
  bytesPerPixel = vi.BytesFromPixels(1);
  vi.SetFieldBased(false);

  // This first release doesn't handle audio, but I still set everything up to
  // at least provide a blank track. At least a few programs I use expect some
  // kind of audio, even if it's silence.
  vi.audio_samples_per_second = 44100;
  vi.nchannels = 1;
  double duration = (double)vi.num_frames /
                    ( (double)vi.fps_numerator / (double)vi.fps_denominator );
  vi.num_audio_samples = (long long)(44100.0 * duration);
  vi.sample_type = SAMPLE_INT16;
  
}



CmvSource::~CmvSource()
{
}



PVideoFrame __stdcall CmvSource::GetFrame(int n, IScriptEnvironment* env)
{

  PVideoFrame dst = env->NewVideoFrame(vi);
  
  const int SET_PITCH = set->GetPitch();
  const unsigned char* setp = set->GetReadPtr();

  const int DST_PITCH = dst->GetPitch();
  unsigned char* dstp = dst->GetWritePtr();

  int tileBytes = tileW * bytesPerPixel;

  int setCols = (set->GetRowSize() / bytesPerPixel) / tileW,
      setRows = set->GetHeight() / tileH;

  int bufFrameOffset = frameSize * n;

  for (int row = 0; row < cmvRows; ++row) {
  
    int dstRow = DST_PITCH * row * tileH;
    int bufRowOffset = bufFrameOffset + (cmvRows - 1 - (row % cmvRows));

    for (int col = 0; col < cmvCols; ++col) {
      
      int curCol = col * tileBytes;
        
      int bufColOffset = bufRowOffset + (col * cmvRows);
      int tileIdx = frameVec[bufColOffset];
      
      unsigned char attribs = attribVec[bufColOffset];

      // With three bits indicating the presence of each component of both
      // the foreground and background colors, pulling them out produces a
      // value from 0-7, which means they can then be used directly as indices
      // to grab the actual packed color from the appropriate vector.
      int fgClrIdx =  attribs & 7,
          bgClrIdx =  (attribs >> 3) & 7,
          fgI =       (attribs >> 6) & 1;

      // That +8 will fall apart if Toady changes the format of colors.txt, but
      // for now it works wonders to separate the dull colors from the bright.
      int bgColor = colors[bgClrIdx],
          fgColor = fgI ? colors[fgClrIdx+8] : colors[fgClrIdx];

      unsigned int  bgClrB = (bgColor & 0x000000FF),
                    bgClrG = (bgColor & 0x0000FF00) >> 8,
                    bgClrR = (bgColor & 0x00FF0000) >> 16;
      
      unsigned int  fgClrB = (fgColor & 0x000000FF),
                    fgClrG = (fgColor & 0x0000FF00) >> 8,
                    fgClrR = (fgColor & 0x00FF0000) >> 16;
      
      // My other plugin, TurnsTile, checks vi.IsRGB() here, but since I'm only
      // working with RGB data for CMVSource I'm free to skip it. RGB is upside
      // down in memory compared to YUV, and I need this to make sure the tile
      // numbering is what you'd expect (top left is 0, top right is 15, etc.).
      int tileIdxY = setRows - 1 - (tileIdx / setCols);
      
      // Modulo here has the effect of "wrapping around" the horizontal tile
      // count for the sheet you've provided.
      int cropLeft = (tileIdx % setCols) * tileBytes,
          cropTop = SET_PITCH * tileIdxY * tileH;
      
      for (int h = 0; h < tileH; ++h) {

        int dstLine = DST_PITCH * h,
            setLine = SET_PITCH * h;

        for (int w = 0; w < tileW; ++w) {
          
          int wBytes = w * bytesPerPixel;

          int dstOffset = dstRow + curCol + dstLine + wBytes,
              setOffset = cropLeft + cropTop + setLine + wBytes;

          int setB = *(setp + setOffset),
              setG = *(setp + setOffset + 1),
              setR = *(setp + setOffset + 2),
              setA = *(setp + setOffset + 3);
          
          int setAInv = 255 - setA;
          
          // Slow, but it works. Somehow, every other trick I've been able to
          // find, wrap my head around and implement for alpha blending has only
          // slowed things down. This is as fast as I can go for the time being.
          if (setA == 0) {
            
            *(dstp + dstOffset) =     (unsigned char)bgClrB;
            *(dstp + dstOffset + 1) = (unsigned char)bgClrG;
            *(dstp + dstOffset + 2) = (unsigned char)bgClrR;
            *(dstp + dstOffset + 3) = 255;

          } else if (setA == 255) {
            
            *(dstp + dstOffset) =     (unsigned char)( (fgClrB * setB) / 255 );
            *(dstp + dstOffset + 1) = (unsigned char)( (fgClrG * setG) / 255 );
            *(dstp + dstOffset + 2) = (unsigned char)( (fgClrR * setR) / 255 );
            *(dstp + dstOffset + 3) = 255;

          } else {
            
            // The two divisions in (((fgClrX * setX) / 255) * setA) / 255 get
            // combined for this; still slow, but every little bit helps.
            *(dstp + dstOffset) =     (unsigned char)( ((fgClrB * setB * setA) / 65025) + ((bgClrB * setAInv) / 255) );
            *(dstp + dstOffset + 1) = (unsigned char)( ((fgClrG * setG * setA) / 65025) + ((bgClrG * setAInv) / 255) );
            *(dstp + dstOffset + 2) = (unsigned char)( ((fgClrR * setR * setA) / 65025) + ((bgClrR * setAInv) / 255) );
            *(dstp + dstOffset + 3) = 255;

          }

        }

      }

    }

  }

  return dst;

}



void __stdcall CmvSource::GetAudio(void* buf, __int64 start, __int64 count, IScriptEnvironment* env)
{

  // As mentioned earlier, at least some of the software I work with regularly
  // expects an audio track, even if it's silence, so I implement GetAudio at
  // just above the bare minimum and zero the track.
  std::memset(buf, 0, static_cast<size_t>(vi.BytesFromAudioSamples(count)));

}